"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var stripAnsi = require("strip-ansi");
var border_1 = require("./border");
var matrix_1 = require("./matrix");
var rectangle_1 = require("./rectangle");
var scroll_bar_1 = require("./scroll-bar");
var utils_1 = require("./utils");
var eaw = require('eastasianwidth');
function defaultSymbolRuler(symbol) {
    var escapedSymbol = stripAnsi(symbol);
    return eaw.characterLength(escapedSymbol);
}
exports.defaultSymbolRuler = defaultSymbolRuler;
;
function createDefaultTrackElement() {
    return matrix_1.createElementBody('|');
}
;
function createDefaultThumbElement() {
    return matrix_1.createElementBody('#');
}
;
function createBox(rectangleLike, options) {
    if (options === void 0) { options = {}; }
    var settings = Object.assign({}, {
        defaultSymbol: null,
    }, options);
    var rectangle = Object.assign({}, { x: 0, y: 0 }, rectangleLike);
    var matrix = matrix_1.createMatrix(rectangle_1.rectangleToSize(rectangle), settings.defaultSymbol);
    if (!matrix_1.validateMatrix(matrix)) {
        throw new Error('The matrix size is invalid');
    }
    return {
        symbolRuler: defaultSymbolRuler,
        x: rectangle.x,
        y: rectangle.y,
        zIndex: 0,
        content: '',
        borders: {
            topWidth: 0,
            bottomWidth: 0,
            leftWidth: 0,
            rightWidth: 0,
            topSymbols: [],
            rightSymbols: [],
            bottomSymbols: [],
            leftSymbols: [],
            topRightSymbols: [],
            bottomRightSymbols: [],
            bottomLeftSymbols: [],
            topLeftSymbols: [],
        },
        matrix: matrix,
        children: [],
    };
}
exports.createBox = createBox;
function applyBoxSettingsToMatrix(box) {
    var pourableElements = matrix_1.parseContent(box.content, box.symbolRuler);
    var filters = [];
    // Borders
    filters.push(function (_a) {
        var matrix = _a.matrix, contentAreaMatrix = _a.contentAreaMatrix, contentArea = _a.contentArea, contentAreaCoordinate = _a.contentAreaCoordinate;
        if (border_1.validateMatrixWithBorders(contentAreaMatrix, box.borders) === false) {
            return {
                matrix: matrix,
                contentArea: contentArea,
                errorOccured: true,
            };
        }
        var placed = border_1.placeBorders(contentAreaMatrix, box.borders);
        return {
            matrix: matrix_1.overwriteMatrix(matrix, placed.matrix, contentAreaCoordinate, box.symbolRuler),
            contentArea: rectangle_1.moveRectangle(placed.contentArea, contentAreaCoordinate),
            errorOccured: false,
        };
    });
    // Scroll Bar
    if (box.scroll !== undefined) {
        var scrollSettings_1 = box.scroll; // To avoid TypeScript's undefined check
        filters.push(function (_a) {
            var matrix = _a.matrix, contentAreaMatrix = _a.contentAreaMatrix, contentArea = _a.contentArea, contentAreaCoordinate = _a.contentAreaCoordinate, contentAreaSize = _a.contentAreaSize;
            if (utils_1.validateSize(contentAreaSize) === false || contentAreaSize.width < 1) {
                return {
                    matrix: matrix,
                    contentArea: contentArea,
                    errorOccured: true,
                };
            }
            var placed = scroll_bar_1.placeScrollBar(contentAreaMatrix, pourableElements, scrollSettings_1.y, scrollSettings_1.trackElement || createDefaultTrackElement(), scrollSettings_1.thumbElement || createDefaultThumbElement());
            return {
                matrix: matrix_1.overwriteMatrix(matrix, placed.matrix, contentAreaCoordinate, box.symbolRuler),
                contentArea: rectangle_1.moveRectangle(placed.contentArea, contentAreaCoordinate),
                errorOccured: false,
            };
        });
    }
    // Content
    if (box.content !== '') {
        var scrollY_1 = box.scroll ? box.scroll.y : 0;
        filters.push(function (_a) {
            var matrix = _a.matrix, contentAreaMatrix = _a.contentAreaMatrix, contentArea = _a.contentArea, contentAreaCoordinate = _a.contentAreaCoordinate;
            var placedMatrix = matrix_1.pourContent(contentAreaMatrix, pourableElements, scrollY_1);
            return {
                matrix: matrix_1.overwriteMatrix(matrix, placedMatrix, contentAreaCoordinate, box.symbolRuler),
                contentArea: contentArea,
                errorOccured: false,
            };
        });
    }
    // Children
    if (box.children.length > 0) {
        filters.push(function (_a) {
            var matrix = _a.matrix, contentAreaMatrix = _a.contentAreaMatrix, contentArea = _a.contentArea, contentAreaCoordinate = _a.contentAreaCoordinate;
            box.children
                .slice()
                .sort(function (a, b) { return a.zIndex - b.zIndex; })
                .forEach(function (childBox) {
                var childMatrix = applyBoxSettingsToMatrix(childBox);
                contentAreaMatrix = matrix_1.overwriteMatrix(contentAreaMatrix, childMatrix, {
                    x: childBox.x,
                    y: childBox.y,
                }, box.symbolRuler);
            });
            return {
                matrix: matrix_1.overwriteMatrix(matrix, contentAreaMatrix, contentAreaCoordinate, box.symbolRuler),
                contentArea: contentArea,
                errorOccured: false,
            };
        });
    }
    var filterResult = filters.reduce(function (preFilterResult, filter) {
        if (preFilterResult.errorOccured) {
            return preFilterResult;
        }
        return filter(Object.assign({
            matrix: preFilterResult.matrix,
            contentArea: preFilterResult.contentArea,
            contentAreaMatrix: matrix_1.cropMatrix(preFilterResult.matrix, preFilterResult.contentArea),
            contentAreaCoordinate: rectangle_1.rectangleToCoordinate(preFilterResult.contentArea),
            contentAreaSize: rectangle_1.rectangleToSize(preFilterResult.contentArea),
        }));
    }, {
        matrix: box.matrix,
        contentArea: matrix_1.matrixToRectangle(box.matrix),
        errorOccured: false,
    });
    return matrix_1.overwriteMatrix(box.matrix, filterResult.matrix, { x: 0, y: 0 }, box.symbolRuler);
}
function renderBox(box, options) {
    if (options === void 0) { options = {}; }
    var settings = Object.assign({
        backgroundSymbol: ' ',
    }, options);
    var matrix = applyBoxSettingsToMatrix(box);
    return matrix_1.renderMatrix(matrix, settings.backgroundSymbol);
}
exports.renderBox = renderBox;
