"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var ansiRegex = require("ansi-regex");
var stripAnsi = require("strip-ansi");
var utils_1 = require("./utils");
var ansiStyles = require('ansi-styles');
var sliceAnsiString = require('slice-ansi-string');
function createDefaultElementStyle() {
    return {
        foregroundColor: '',
        backgroundColor: '',
        bold: false,
        dim: false,
        italic: false,
        underline: false,
        inverse: false,
        hidden: false,
        strikethrough: false,
    };
}
exports.createDefaultElementStyle = createDefaultElementStyle;
function createElementBody(symbol) {
    return {
        symbol: symbol,
        style: createDefaultElementStyle(),
    };
}
exports.createElementBody = createElementBody;
function createElement(y, x, symbol) {
    return {
        y: y,
        x: x,
        symbol: symbol,
        style: createDefaultElementStyle(),
    };
}
function createMatrix(size, defaultSymbol) {
    if (defaultSymbol === void 0) { defaultSymbol = null; }
    var matrix = [];
    for (var y = 0; y < size.height; y += 1) {
        var row = [];
        for (var x = 0; x < size.width; x += 1) {
            row.push(createElement(y, x, defaultSymbol));
        }
        matrix.push(row);
    }
    return matrix;
}
exports.createMatrix = createMatrix;
/**
 * Validate that the matrix is not empty and is rectangular
 */
function validateMatrix(matrix) {
    return (Array.isArray(matrix) &&
        matrix.length > 0 &&
        Array.isArray(matrix[0]) &&
        matrix[0].length > 0 &&
        matrix.every(function (row) { return row.length === matrix[0].length; }));
}
exports.validateMatrix = validateMatrix;
function getElement(matrix, coordinate) {
    var row = matrix[coordinate.y];
    if (!row) {
        return null;
    }
    return row[coordinate.x] || null;
}
exports.getElement = getElement;
/**
 * A function that makes it easy to create a matrix from text mainly for testing.
 *
 * @param text Text consisting only of ASCII characters.
 *             It also needs to be rectangular.
 */
function createMatrixFromText(text) {
    function parseTextToSymbols(text) {
        return text
            .replace(/\n+$/, '')
            .split('\n')
            .map(function (row) { return row.split(''); });
    }
    var symbols = parseTextToSymbols(text);
    var matrix = createMatrix({
        width: symbols[0].length,
        height: symbols.length,
    });
    for (var y = 0; y < symbols.length; y += 1) {
        for (var x = 0; x < symbols[0].length; x += 1) {
            var element = getElement(matrix, { x: x, y: y });
            if (element) {
                element.symbol = symbols[y][x];
            }
            else {
                throw new Error('The `element` must exist.');
            }
        }
    }
    return matrix;
}
exports.createMatrixFromText = createMatrixFromText;
function getWidth(matrix) {
    if (matrix.length === 0) {
        return 0;
    }
    return matrix[0].length;
}
exports.getWidth = getWidth;
function getHeight(matrix) {
    return matrix.length;
}
exports.getHeight = getHeight;
function getMaxX(matrix) {
    return getWidth(matrix) - 1;
}
exports.getMaxX = getMaxX;
function getMaxY(matrix) {
    return getHeight(matrix) - 1;
}
exports.getMaxY = getMaxY;
function matrixToRectangle(matrix) {
    return {
        x: 0,
        y: 0,
        width: getWidth(matrix),
        height: getHeight(matrix),
    };
}
exports.matrixToRectangle = matrixToRectangle;
// TODO: Negative coordinates
function overwriteMatrix(matrix, replacer, coordinate, symbolRuler) {
    if (getWidth(replacer) === 0 || getHeight(replacer) === 0) {
        return matrix;
    }
    var newMatrix = matrix.map(function (row, y) {
        return row.map(function (element, x) {
            if (y >= coordinate.y && y < coordinate.y + getHeight(replacer) &&
                x >= coordinate.x && x < coordinate.x + getWidth(replacer)) {
                var newElement = getElement(replacer, {
                    y: y - coordinate.y,
                    x: x - coordinate.x,
                });
                if (newElement) {
                    return newElement;
                }
                else {
                    throw new Error('The `newElement` must exist.');
                }
            }
            return element;
        });
    });
    //
    // Delete separated multibyte fragments in following cases.
    //
    // 1) A case that a multibyte fragment exist on the left edge of the replacer.
    //
    //     (e.g.)  | first x - 1 | first x | <- replacer's position
    //    ---------+-------------+---------+
    //    original |     "a"     |   "b"   |
    //    replacer |    (None)   |  false  |
    //    current  |     "a"     |  false  |
    //    fix      |     "a"     |   null  |
    //
    // 2) A case that a multibyte fragment exist on the right edge of the replacer.
    //
    //     (e.g.)  |   last x   | last.x + 1 | <- replacer's position
    //    ---------+------------+------------+
    //    original |     "a"    |     "b"    |
    //    replacer |  multibyte |   (None)   |
    //    current  |  multibyte |     "b"    |
    //    fix      |     null   |     "b"    |
    //
    // 3) A case that a multibyte fragment exist on the left side of the replacer.
    //
    //     (e.g.)  | first x - 1 | first x | <- replacer's position
    //    ---------+-------------+---------+
    //    original |  multibyte  |  false  |
    //    replacer |    (None)   |   "a"   |
    //    current  |  multibyte  |   "a"   |
    //    fix      |     null    |   "a"   |
    //
    // 4) A case that a multibyte fragment exist on the right side of the replacer.
    //
    //     (e.g.)  |   last x   | last.x + 1 | <- replacer's position
    //    ---------+------------+------------+
    //    original |  multibyte |   false    |
    //    replacer |     "a"    |   (None)   |
    //    current  |     "a"    |   false    |
    //    fix      |     "a"    |    null    |
    //
    // Complex cases)
    //
    //     (e.g.)  | first x - 1 | first x | <- replacer's position
    //    ---------+-------------+---------+
    //    original |  multibyte  |  false  |
    //    replacer |    (None)   |  false  |
    //    current  |  multibyte  |  false  |
    //    fix in 1 |  multibyte  |   null  |
    //    fix in 3 |     null    |   null  |
    //
    //     (e.g.)  |   last x   | last.x + 1 | <- replacer's position
    //    ---------+------------+------------+
    //    original |  multibyte |   false    |
    //    replacer |  multibyte |   (None)   |
    //    current  |  multibyte |   false    |
    //    fix in 2 |     null   |   false    |
    //    fix in 4 |     null   |    null    |
    //
    //    These cases are somewhat unnatural because they eliminate multibyte characters that need not be erased.
    //    However, these situations are edge cases because basically the `replacer` often has borders.
    //
    newMatrix = newMatrix.map(function (row, y) {
        return row.map(function (element, x) {
            // 1)
            if (x === coordinate.x &&
                element.symbol === false) {
                return Object.assign({}, element, { symbol: null });
                // 2)
            }
            else if (x === coordinate.x + getWidth(replacer) - 1 &&
                typeof element.symbol === 'string' &&
                symbolRuler(element.symbol) === 2) {
                return Object.assign({}, element, { symbol: null });
                // 3)
            }
            else if (x === coordinate.x - 1 &&
                typeof element.symbol === 'string' &&
                symbolRuler(element.symbol) === 2) {
                return Object.assign({}, element, { symbol: null });
                // 4)
            }
            else if (x === coordinate.x + getWidth(replacer) &&
                element.symbol === false) {
                return Object.assign({}, element, { symbol: null });
            }
            return element;
        });
    });
    return newMatrix;
}
exports.overwriteMatrix = overwriteMatrix;
function cropMatrix(matrix, rectangle) {
    var newMatrix = [];
    for (var y = rectangle.y; y < rectangle.y + rectangle.height; y += 1) {
        var row = [];
        for (var x = rectangle.x; x < rectangle.x + rectangle.width; x += 1) {
            var element = getElement(matrix, { x: x, y: y });
            if (element) {
                row.push(element);
            }
        }
        if (row.length > 0) {
            newMatrix.push(row);
        }
    }
    return newMatrix;
}
exports.cropMatrix = cropMatrix;
/**
 * @param character A single character with or without ANSI escape codes.
 *                  It is mainly assumed to pass the return value of "slice-ansi-string".
 */
function decodeAnsiStyles(character) {
    var characterWithoutAnsi = stripAnsi(character);
    if (utils_1.stringToArray(characterWithoutAnsi).length !== 1) {
        throw new Error('It should be used for single characters.');
    }
    var ansiEscapeCodes = character.match(ansiRegex()) || [];
    var styleData = {};
    // From https://github.com/chalk/ansi-styles#colors
    var foregroundColorNames = [
        'black',
        'red',
        'green',
        'yellow',
        'blue',
        'magenta',
        'cyan',
        'white',
        // NOTE: "blackBright" will be added in the following commit.
        //       https://github.com/chalk/ansi-styles/commit/fb5b656d9fce745881c36deb8ff800b5080d9f89
        'grey',
        //'blackBright',
        'redBright',
        'greenBright',
        'yellowBright',
        'blueBright',
        'magentaBright',
        'cyanBright',
        'whiteBright',
    ];
    ansiEscapeCodes.some(function (code) {
        var foundName = foregroundColorNames.find(function (name) {
            return code === (ansiStyles.color[name] && ansiStyles.color[name].open);
        });
        if (foundName) {
            styleData.foregroundColor = foundName;
            return true;
        }
        return false;
    });
    // From https://github.com/chalk/ansi-styles#background-colors
    var backgroundColorNames = [
        'bgBlack',
        'bgRed',
        'bgGreen',
        'bgYellow',
        'bgBlue',
        'bgMagenta',
        'bgCyan',
        'bgWhite',
        'bgBlackBright',
        'bgRedBright',
        'bgGreenBright',
        'bgYellowBright',
        'bgBlueBright',
        'bgMagentaBright',
        'bgCyanBright',
        'bgWhiteBright',
    ];
    ansiEscapeCodes.some(function (code) {
        var foundName = backgroundColorNames.find(function (name) {
            return code === (ansiStyles.bgColor[name] && ansiStyles.bgColor[name].open);
        });
        if (foundName) {
            styleData.backgroundColor = foundName;
            return true;
        }
        return false;
    });
    // From) https://github.com/chalk/ansi-styles#modifiers
    ansiEscapeCodes.forEach(function (code) {
        if (code === ansiStyles.bold.open) {
            styleData.bold = true;
        }
        else if (code === ansiStyles.dim.open) {
            styleData.dim = true;
        }
        else if (code === ansiStyles.italic.open) {
            styleData.italic = true;
        }
        else if (code === ansiStyles.underline.open) {
            styleData.underline = true;
        }
        else if (code === ansiStyles.inverse.open) {
            styleData.inverse = true;
        }
        else if (code === ansiStyles.hidden.open) {
            styleData.hidden = true;
        }
        else if (code === ansiStyles.strikethrough.open) {
            styleData.strikethrough = true;
        }
    });
    return styleData;
}
exports.decodeAnsiStyles = decodeAnsiStyles;
function parseContent(content, symbolRuler) {
    var pourableElements = [];
    // This pointer is considering ANSI escape code.
    // Therefore, it can not be used with `str.slice()` and so on.
    var pointer = 0;
    while (true) {
        var symbolWithAnsi = sliceAnsiString(content, pointer, pointer + 1);
        if (symbolWithAnsi === '') {
            break;
        }
        var symbol = stripAnsi(symbolWithAnsi);
        var symbolWidth = symbolRuler(symbol);
        if (symbol === '\n') {
            pourableElements.push({
                isLineBreaking: true,
                symbol: '',
                symbolWidth: symbolWidth,
                style: createDefaultElementStyle(),
            });
        }
        else {
            pourableElements.push({
                isLineBreaking: false,
                symbol: symbol,
                symbolWidth: symbolWidth,
                style: Object.assign(createDefaultElementStyle(), decodeAnsiStyles(symbolWithAnsi)),
            });
        }
        pointer += 1;
    }
    return pourableElements;
}
exports.parseContent = parseContent;
/**
 * Try to pour elements to a certain width
 */
function pourElementsVirtually(pourableElements, matrixWidth) {
    var newPourableElements = [];
    var isFinished = false;
    var yPointer = 0;
    var xPointer = 0;
    pourableElements.forEach(function (pourableElement) {
        if (isFinished) {
            return;
        }
        if (pourableElement.isLineBreaking) {
            yPointer += 1;
            xPointer = 0;
        }
        else if (pourableElement.symbolWidth === 2 && matrixWidth === 1) {
            // The content can not pour anymore.
            isFinished = true;
        }
        else {
            if (pourableElement.symbolWidth === 2) {
                if (
                // |abcd|P => |abcd|
                // |....|     |P...|
                // (P = pointer)
                xPointer === matrixWidth ||
                    // |abcP| => |abcd|
                    // |....|    |P...|
                    // (P = pointer)
                    xPointer + 1 === matrixWidth) {
                    yPointer += 1;
                    xPointer = 0;
                }
                newPourableElements.push(Object.assign({}, pourableElement, {
                    x: xPointer,
                    y: yPointer,
                }));
                xPointer += 1;
                newPourableElements.push(Object.assign({}, pourableElement, {
                    symbol: false,
                    x: xPointer,
                    y: yPointer,
                }));
                xPointer += 1;
            }
            else if (pourableElement.symbolWidth === 1) {
                if (xPointer === matrixWidth) {
                    yPointer += 1;
                    xPointer = 0;
                }
                newPourableElements.push(Object.assign({}, pourableElement, {
                    x: xPointer,
                    y: yPointer,
                }));
                xPointer += 1;
            }
        }
    });
    return {
        pouredElements: newPourableElements,
        // No content is regarded as zero height.
        contentHeight: (yPointer === 0 && xPointer === 0) ? 0 : yPointer + 1,
    };
}
exports.pourElementsVirtually = pourElementsVirtually;
function pourContent(matrix, pourableElements, startingY) {
    var width = getWidth(matrix);
    var height = getHeight(matrix);
    var maxY = getMaxY(matrix);
    // Reset the matrix to background only.
    var newMatrix = createMatrix({ width: width, height: height }, null);
    var pouredElements = pourElementsVirtually(pourableElements, width).pouredElements;
    pouredElements.forEach(function (pouredElement) {
        var y = pouredElement.y - startingY;
        if (y < 0 || y > maxY) {
            return;
        }
        Object.assign(newMatrix[y][pouredElement.x], {
            symbol: pouredElement.symbol,
            style: pouredElement.style,
        });
    });
    return newMatrix;
}
exports.pourContent = pourContent;
function renderElement(element) {
    var symbol = element.symbol, style = element.style;
    var modifiers = [];
    if (style.foregroundColor) {
        modifiers.push(ansiStyles.color[style.foregroundColor].open);
    }
    if (style.backgroundColor) {
        modifiers.push(ansiStyles.bgColor[style.backgroundColor].open);
    }
    if (style.bold) {
        modifiers.push(ansiStyles.bold.open);
    }
    if (style.dim) {
        modifiers.push(ansiStyles.dim.open);
    }
    if (style.italic) {
        modifiers.push(ansiStyles.italic.open);
    }
    if (style.underline) {
        modifiers.push(ansiStyles.underline.open);
    }
    if (style.inverse) {
        modifiers.push(ansiStyles.inverse.open);
    }
    if (style.hidden) {
        modifiers.push(ansiStyles.hidden.open);
    }
    if (style.strikethrough) {
        modifiers.push(ansiStyles.strikethrough.open);
    }
    return modifiers.join('') + symbol + (modifiers.length > 0 ? ansiStyles.reset.close : '');
}
function renderMatrix(matrix, backgroundSymbol) {
    return matrix
        .map(function (row) {
        return row.map(function (element) {
            if (element.symbol === false) {
                return '';
            }
            else if (element.symbol === null) {
                return backgroundSymbol;
            }
            return renderElement(element);
        }).join('');
    })
        .join('\n');
}
exports.renderMatrix = renderMatrix;
;
